@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # This file is part of Xpra.
-# Copyright (C) 2014-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2014-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -15,8 +15,9 @@ from xpra.gtk_common.gtk_util import (
     choose_file, get_gtk_version_info,
     JUSTIFY_LEFT, WIN_POS_CENTER, FILE_CHOOSER_ACTION_SAVE,
     )
+from xpra.platform.gui import force_focus
 from xpra.util import nonl, envint, repr_ellipsized
-from xpra.os_util import strtobytes
+from xpra.os_util import strtobytes, hexstr, PYTHON3
 from xpra.log import Logger
 
 log = Logger("util")
@@ -30,17 +31,17 @@ STEP_DELAY = envint("XPRA_BUG_REPORT_STEP_DELAY", 0)
 
 class BugReport(object):
 
-    def init(self, show_about=True, get_server_info=None, opengl_info=None, includes={}):
+    def init(self, show_about=True, get_server_info=None, opengl_info=None, includes=None):
         self.show_about = show_about
         self.get_server_info = get_server_info
         self.opengl_info = opengl_info
-        self.includes = includes
+        self.includes = includes or {}
         self.setup_window()
 
     def setup_window(self):
         self.window = gtk.Window()
         window_defaults(self.window)
-        self.window.connect("destroy", self.close)
+        self.window.connect("delete-event", self.close)
         self.window.set_default_size(400, 300)
         self.window.set_title("Xpra Bug Report")
 
@@ -103,7 +104,7 @@ class BugReport(object):
             from xpra.sound.wrapper import query_sound
             def get_sound_info():
                 return query_sound()
-        except:
+        except ImportError:
             get_sound_info = None
         def get_gl_info():
             if self.opengl_info:
@@ -141,16 +142,16 @@ class BugReport(object):
         try:
             from xpra.platform.gui import take_screenshot
             take_screenshot_fn = take_screenshot
-        except:
+        except ImportError:
             log("failed to load platfrom specific screenshot code", exc_info=True)
         if not take_screenshot_fn:
             #try with Pillow:
             try:
                 from PIL import ImageGrab           #@UnresolvedImport
-                from xpra.os_util import StringIOClass
+                from io import BytesIO
                 def pillow_imagegrab_screenshot():
                     img = ImageGrab.grab()
-                    out = StringIOClass()
+                    out = BytesIO()
                     img.save(out, format="PNG")
                     v = out.getvalue()
                     out.close()
@@ -164,13 +165,14 @@ class BugReport(object):
                 from xpra.server.shadow.gtk_root_window_model import GTKImageCapture
                 rwm = GTKImageCapture(get_default_root_window())
                 take_screenshot_fn = rwm.take_screenshot
-            except:
+            except Exception:
                 log.warn("Warning: failed to load gtk screenshot code", exc_info=True)
         log("take_screenshot_fn=%s", take_screenshot_fn)
         if take_screenshot_fn:
-            def get_screenshot():
+            def _get_screenshot():
                 #take_screenshot() returns: w, h, "png", rowstride, data
                 return take_screenshot_fn()[4]
+            get_screenshot = _get_screenshot
         self.toggles = (
                    ("system",       "txt",  "System",           get_sys_info,
                     "Xpra version, platform and host information"),
@@ -211,9 +213,7 @@ class BugReport(object):
             hbox.pack_start(btn)
             return btn
 
-        if not is_gtk3():
-            #clipboard does not work in gtk3..
-            btn("Copy to clipboard", "Copy all data to clipboard", self.copy_clicked, "clipboard.png")
+        btn("Copy to clipboard", "Copy all data to clipboard", self.copy_clicked, "clipboard.png")
         btn("Save", "Save Bug Report", self.save_clicked, "download.png")
         btn("Cancel", "", self.close, "quit.png")
 
@@ -229,17 +229,21 @@ class BugReport(object):
         log("show()")
         if not self.window:
             self.setup_window()
+        force_focus()
         self.window.show_all()
         self.window.present()
 
     def hide(self):
         log("hide()")
-        self.window.hide()
+        if self.window:
+            self.window.hide()
 
     def close(self, *args):
         log("close%s", args)
-        self.hide()
-        self.window = None
+        if self.window:
+            self.hide()
+            self.window = None
+        return True
 
     def destroy(self, *args):
         log("destroy%s", args)
@@ -267,8 +271,8 @@ class BugReport(object):
         return None
 
 
-    def get_text_data(self):
-        log("get_text_data() collecting bug report data")
+    def get_data(self):
+        log("get_data() collecting bug report data")
         data = []
         tb = self.description.get_buffer()
         buf = tb.get_text(*tb.get_bounds(), include_hidden_chars=False)
@@ -302,15 +306,19 @@ class BugReport(object):
             elif isinstance(value, (list, tuple)):
                 s = os.linesep.join(str(x) for x in value)
             else:
-                s = str(value)
+                s = value
             log("%s (%s) %s: %s", title, tooltip, dtype, repr_ellipsized(s))
             data.append((title, tooltip, dtype, s))
             time.sleep(STEP_DELAY)
         return data
 
     def copy_clicked(self, *_args):
-        data = self.get_text_data()
-        text = os.linesep.join("%s: %s%s%s%s" % (title, tooltip, os.linesep, v, os.linesep)
+        data = self.get_data()
+        def cdata(v):
+            if isinstance(v, bytes):
+                return hexstr(v)
+            return str(v)
+        text = os.linesep.join("%s: %s%s%s%s" % (title, tooltip, os.linesep, cdata(v), os.linesep)
                                for (title,tooltip,dtype,v) in data if dtype=="txt")
         clipboard = gtk.clipboard_get(gdk.SELECTION_CLIPBOARD)
         clipboard.set_text(text)
@@ -327,7 +335,7 @@ class BugReport(object):
         if not filename.lower().endswith(".zip"):
             filename = filename+".zip"
         basenoext = os.path.splitext(os.path.basename(filename))[0]
-        data = self.get_text_data()
+        data = self.get_data()
         import zipfile
         zf = zipfile.ZipFile(filename, mode='w', compression=zipfile.ZIP_DEFLATED)
         try:
@@ -338,6 +346,23 @@ class BugReport(object):
                 #very poorly documented:
                 info.external_attr = 0o644 << 16
                 info.comment = strtobytes(tooltip)
-                zf.writestr(info, s)
+                if PYTHON3 and isinstance(s, bytes):
+                    try:
+                        try:
+                            import tempfile
+                            temp = tempfile.NamedTemporaryFile(prefix="xpra.", suffix=".screenshot", delete=False)
+                            with temp:
+                                temp.write(s)
+                                temp.flush()
+                        except OSError as e:
+                            log.error("Error: cannot create mmap file:")
+                            log.error(" %s", e)
+                        else:
+                            zf.write(temp.name, cfile, zipfile.ZIP_STORED)
+                    finally:
+                        if temp:
+                            os.unlink(temp.name)
+                else:
+                    zf.writestr(info, str(s))
         finally:
             zf.close()
