@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2017 Antoine Martin <antoine@devloop.org.uk>
+ * Copyright (c) 2013-2017 Antoine Martin <antoine@xpra.org>
  * Copyright (c) 2014 Joshua Higgins <josh@kxes.net>
  * Copyright (c) 2015-2016 Spikes, Inc.
  * Licensed under MPL 2.0
@@ -28,6 +28,7 @@ function XpraWindow(client, canvas_state, wid, x, y, w, h, metadata, override_re
 	this.log = function() { client.log.apply(client, arguments); };
 	this.warn = function() { client.warn.apply(client, arguments); };
 	this.error = function() { client.error.apply(client, arguments); };
+	this.exc = function() { client.exc.apply(client, arguments); };
 	this.debug = function() { client.debug.apply(client, arguments); };
 	this.debug_categories = client.debug_categories;
 	//keep reference both the internal canvas and screen drawn canvas:
@@ -112,7 +113,7 @@ function XpraWindow(client, canvas_state, wid, x, y, w, h, metadata, override_re
 		jQuery(this.div).addClass("window-" + this.windowtype);
 	}
 
-	if (this.client.server_is_desktop) {
+	if (this.client.server_is_desktop || this.client.server_is_shadow) {
 		jQuery(this.div).addClass("desktop");
 		this.resizable = false;
 	}
@@ -124,21 +125,26 @@ function XpraWindow(client, canvas_state, wid, x, y, w, h, metadata, override_re
 	}
 	else if((this.windowtype == "") || (this.windowtype == "NORMAL") || (this.windowtype == "DIALOG") || (this.windowtype == "UTILITY")) {
 		this.resizable = true;
+		jQuery(this.div).addClass("border");
 		// add a title bar to this window if we need to
 		// create header
 		jQuery(this.div).prepend('<div id="head' + String(wid) + '" class="windowhead"> '+
 				'<span class="windowicon"><img id="windowicon' + String(wid) + '" /></span> '+
 				'<span class="windowtitle" id="title' + String(wid) + '">' + this.title + '</span> '+
 				'<span class="windowbuttons"> '+
+				'<span id="minimize' + String(wid) + '"><img src="icons/minimize.png" /></span> '+
 				'<span id="maximize' + String(wid) + '"><img src="icons/maximize.png" /></span> '+
 				'<span id="close' + String(wid) + '"><img src="icons/close.png" /></span> '+
 				'</span></div>');
 		// make draggable
 		jQuery(this.div).draggable({ cancel: "canvas" });
 		jQuery(this.div).on("dragstart",function(ev,ui){
-			set_focus_cb(me);
+			client.do_window_mouse_click(ev, me, false);
+			client.mouse_grabbed = true;
+			me.set_focus_cb(me);
 		});
 		jQuery(this.div).on("dragstop",function(ev,ui){
+			client.mouse_grabbed = false;
 			me.handle_moved(ui);
 		});
 		// attach resize handles
@@ -146,30 +152,58 @@ function XpraWindow(client, canvas_state, wid, x, y, w, h, metadata, override_re
 		//jQuery(this.div).on("resize",jQuery.debounce(50, function(ev,ui) {
 		//  	me.handle_resized(ui);
 		//}));
+		jQuery(this.div).on("resizestart",function(ev,ui){
+			client.do_window_mouse_click(ev, me, false);
+			client.mouse_grabbed = true;
+		});
 		jQuery(this.div).on("resizestop",function(ev,ui){
 		  	me.handle_resized(ui);
-		  	set_focus_cb(me);
+		  	me.set_focus_cb(me);
+			client.mouse_grabbed = false;
+			//workaround for the window going blank,
+			//just force a refresh:
+			setTimeout(function() {
+				me.client.request_refresh(me.wid);
+			}, 200);
 		});
 		this.d_header = '#head' + String(wid);
 		this.d_closebtn = '#close' + String(wid);
 		this.d_maximizebtn = '#maximize' + String(wid);
+		this.d_minimizebtn = '#minimize' + String(wid);
 		if (this.resizable) {
+			jQuery(this.d_header).dblclick(function() {
+				me.toggle_maximized();
+			});
 			jQuery(this.d_closebtn).click(function() {
 				window_closed_cb(me);
 			});
 			jQuery(this.d_maximizebtn).click(function() {
 				me.toggle_maximized();
 			});
+			jQuery(this.d_minimizebtn).click(function() {
+				me.toggle_minimized();
+			});
 		}
 		else {
 			jQuery(this.d_closebtn).hide();
 			jQuery(this.d_maximizebtn).hide();
+			jQuery('#windowlistitemmax' + String(wid)).hide();
+			jQuery(this.d_minimizebtn).hide();
 		}
 		// adjust top offset
 		this.topoffset = this.topoffset + parseInt(jQuery(this.d_header).css('height'), 10);
+		// stop propagation if we're over the window:
+		jQuery(this.div).mousedown(function (e) {
+			e.stopPropagation();
+		});
+		//bug 2418: if we stop 'mouseup' propagation,
+		//jQuery can't ungrab the window with Firefox
+		//jQuery(this.div).mouseup(function (e) {
+		//	e.stopPropagation();
+		//});
 		// assign some interesting callbacks
 		jQuery(this.d_header).click(function() {
-			set_focus_cb(me);
+			me.set_focus_cb(me);
 		});
 	}
 
@@ -192,11 +226,48 @@ function XpraWindow(client, canvas_state, wid, x, y, w, h, metadata, override_re
 		div.addEventListener('DOMMouseScroll',	on_mousescroll, false); // for Firefox
 	}
 
-	// need to update the CSS geometry
-	this.ensure_visible();
-	this.updateCSSGeometry();
-	//show("placing new window at "+this.x+","+this.y);
+	this.pointer_down = -1;
+	this.pointer_last_x = 0;
+	this.pointer_last_y = 0;
+	if (window.PointerEvent) {
+		this.canvas.addEventListener("pointerdown", function(ev) {
+			me.debug("mouse", "pointerdown:", ev);
+			if (ev.pointerType=="touch") {
+				me.pointer_down = ev.pointerId;
+				me.pointer_last_x = ev.offsetX;
+				me.pointer_last_y = ev.offsetY;
+			}
+		});
+		this.canvas.addEventListener("pointermove", function(ev) {
+			me.debug("mouse", "pointermove:", ev);
+			if (me.pointer_down==ev.pointerId) {
+				var dx = ev.offsetX-me.pointer_last_x;
+				var dy = ev.offsetY-me.pointer_last_y;
+				me.pointer_last_x = ev.offsetX;
+				me.pointer_last_y = ev.offsetY;
+				var mult = 20.0*(window.devicePixelRatio || 1);
+				ev.wheelDeltaX = Math.round(dx*mult);
+				ev.wheelDeltaY = Math.round(dy*mult);
+				on_mousescroll(ev);
+			}
+		});
+		this.canvas.addEventListener("pointerup", function(ev) {
+			me.debug("mouse", "pointerup:", ev);
+			me.pointer_down = -1;
+		});
+		this.canvas.addEventListener("pointercancel", function(ev) {
+			me.debug("mouse", "pointercancel:", ev);
+			me.pointer_down = -1;
+		});
+		this.canvas.addEventListener("pointerout", function(ev) {
+			me.debug("mouse", "pointerout:", ev);
+		});
+	}
 
+	// adapt to screen size if needed (ie: shadow / desktop windows):
+	this.screen_resized();
+	// set the CSS geometry
+	this.updateCSSGeometry();
 	// now read all metadata
 	this.update_metadata(metadata);
 };
@@ -210,9 +281,10 @@ XpraWindow.prototype._init_2d_canvas = function() {
 
 XpraWindow.prototype.swap_buffers = function() {
 	//the up to date canvas is what we'll draw on screen:
+	this.debug("draw", "swap_buffers");
 	this.draw_canvas = this.offscreen_canvas;
 	this._init_2d_canvas();
-	this.offscreen_canvas_ctx.drawImage(this.draw_canvas, 0, 0);	
+	this.offscreen_canvas_ctx.drawImage(this.draw_canvas, 0, 0);
 }
 
 XpraWindow.prototype.set_spinner = function(state) {
@@ -224,6 +296,11 @@ XpraWindow.prototype.set_spinner = function(state) {
 }
 
 XpraWindow.prototype.ensure_visible = function() {
+	if (this.client.server_is_desktop || this.client.server_is_shadow) {
+		//those windows should usually be centered on screen,
+		//moving them would mess that up
+		return;
+	}
 	var oldx = this.x;
 	var oldy = this.y;
 	// for now make sure we don't out of top left
@@ -245,6 +322,7 @@ XpraWindow.prototype.ensure_visible = function() {
 	else if (oldy >= wh - min_visible) {
 		this.y = Math.min(oldy, wh - min_visible);
 	}
+	this.debug("geometry", "ensure_visible() oldx=", oldx, "oldy=", oldy, "x=", this.x, "y=", this.y);
 	if((oldx != this.x) || (oldy != this.y)) {
 		this.updateCSSGeometry();
 		return false;
@@ -271,7 +349,7 @@ XpraWindow.prototype.updateCanvasGeometry = function() {
 XpraWindow.prototype.updateCSSGeometry = function() {
 	// set size of canvas
 	this.updateCanvasGeometry();
-	if (this.client.server_is_desktop) {
+	if (this.client.server_is_desktop || this.client.server_is_shadow) {
 		jQuery(this.div).position({of : jQuery("#screen")});
 		return;
 	}
@@ -286,6 +364,7 @@ XpraWindow.prototype.updateCSSGeometry = function() {
 	this.outerY = this.y - this.topoffset;
 	jQuery(this.div).css('left', this.outerX);
 	jQuery(this.div).css('top', this.outerY);
+	this.debug("geometry", "updateCSSGeometry() left=", this.outerX, ", top=", this.outerY, ", width=", this.outerW, ", height=", this.outerH);
 }
 
 XpraWindow.prototype.updateFocus = function() {
@@ -322,7 +401,7 @@ XpraWindow.prototype.on_mouseup = function(e) {
 
 XpraWindow.prototype.on_mousescroll = function(e) {
 	this.mouse_scroll_cb(this.client, e, this);
-	e.preventDefault();
+	//e.preventDefault();
 	return false;
 }
 
@@ -339,8 +418,8 @@ XpraWindow.prototype.update_zindex = function() {
 	if (this.tray) {
 		z = 0;
 	}
-	else if (this.override_redirect || this.client.server_is_desktop) {
-		z = 15000;
+	else if (this.override_redirect || this.client.server_is_desktop || this.client.server_is_shadow) {
+		z = 30000;
 	}
 	else if (this.windowtype=="DROPDOWN" || this.windowtype=="TOOLTIP" ||
 			this.windowtype=="POPUP_MENU" || this.windowtype=="MENU" ||
@@ -391,7 +470,10 @@ XpraWindow.prototype.update_metadata = function(metadata, safe) {
 XpraWindow.prototype.set_metadata_safe = function(metadata) {
 	if ("title" in metadata) {
 		this.title = metadata["title"];
-		jQuery('#title' + this.wid).html(decodeURIComponent(escape(this.title)));
+		var decodedTitle = decodeURIComponent(escape(this.title));
+		jQuery('#title' + this.wid).html(decodedTitle);
+        var trimmedTitle = Utilities.trimString(decodedTitle,30);
+		jQuery('#windowlistitemtitle'+this.wid).text(trimmedTitle);
 	}
 	if ("has-alpha" in metadata) {
 		this.has_alpha = metadata["has-alpha"];
@@ -412,7 +494,7 @@ XpraWindow.prototype.set_metadata_safe = function(metadata) {
 			opacity = 1.0;
 		}
 		else {
-			opacity = opacity / 0x100000000
+			opacity = opacity / 0x100000000;
 		}
 		jQuery(this.div).css('opacity', ''+opacity);
 	}
@@ -458,7 +540,6 @@ XpraWindow.prototype.set_metadata_safe = function(metadata) {
 };
 
 XpraWindow.prototype.apply_size_constraints = function() {
-	var size_constraints = this.metadata["size-constraints"];
 	if (!this.resizable) {
 		return;
 	}
@@ -474,6 +555,7 @@ XpraWindow.prototype.apply_size_constraints = function() {
 		hdec = jQuery('#head' + this.wid).outerHeight(true);
 	}
 	var min_size = null, max_size = null;
+	var size_constraints = this.metadata["size-constraints"];
 	if (size_constraints) {
 		min_size = size_constraints["minimum-size"];
 		max_size = size_constraints["maximum-size"];
@@ -490,6 +572,7 @@ XpraWindow.prototype.apply_size_constraints = function() {
 	}
 	if(minw>0 && minw==maxw && minh>0 && minh==maxh) {
 		jQuery(this.d_maximizebtn).hide();
+		jQuery('#windowlistitemmax' + String(this.wid)).hide();
 		jQuery(this.div).resizable('disable');
 	} else {
 		jQuery(this.d_maximizebtn).show();
@@ -533,6 +616,7 @@ XpraWindow.prototype.save_geometry = function() {
 			"y"	: this.y,
 			"w"	: this.w,
 			"h" : this.h};
+	this.debug("geometry", "save_geometry() saved-geometry=", this.saved_geometry);
 }
 /**
  * Restores the saved geometry (if it exists).
@@ -545,6 +629,7 @@ XpraWindow.prototype.restore_geometry = function() {
 	this.y = this.saved_geometry["y"];
 	this.w = this.saved_geometry["w"];
 	this.h = this.saved_geometry["h"];
+	this.debug("geometry", "restore_geometry() saved-geometry=", this.saved_geometry);
 	// delete saved geometry
 	this.saved_geometry = null;
 	// then call local resized callback
@@ -556,6 +641,10 @@ XpraWindow.prototype.restore_geometry = function() {
  * Maximize / unmaximizes the window.
  */
 XpraWindow.prototype.set_maximized = function(maximized) {
+	if(jQuery(this.div).is(":hidden")){
+		jQuery(this.div).show();
+	}
+
 	if (this.maximized==maximized) {
 		return;
 	}
@@ -574,6 +663,21 @@ XpraWindow.prototype.toggle_maximized = function() {
 	this.set_maximized(!this.maximized);
 };
 
+/**
+ * Minimizes / unminimizes the window.
+ */
+XpraWindow.prototype.set_minimized = function(minimized) {
+	jQuery(this.div).toggle(200);
+};
+
+
+/**
+ * Toggle minimized state
+ */
+XpraWindow.prototype.toggle_minimized = function() {
+	this.set_minimized(!this.minimized);
+};
+
 /**
  * Fullscreen / unfullscreen the window.
  */
@@ -610,6 +714,7 @@ XpraWindow.prototype._set_decorated = function(decorated) {
 		jQuery(this.div).addClass("window");
 		if (this.d_header) {
 			this.topoffset = this.topoffset + parseInt(jQuery(this.d_header).css('height'), 10);
+			this.debug("geometry", "_set_decorated(", decorated, ") new topoffset=", self.topoffset);
 		}
 	}
 	else {
@@ -646,6 +751,7 @@ XpraWindow.prototype.fill_screen = function() {
 	this.y = 0 + this.topoffset;
 	this.w = (screen_size[0] - this.leftoffset) - this.rightoffset;
 	this.h = (screen_size[1] - this.topoffset) - this.bottomoffset;
+	this.debug("geometry", "fill_screen() ", this.x, this.y, this.w, this.h);
 };
 
 
@@ -660,6 +766,7 @@ XpraWindow.prototype.handle_resized = function(e) {
 	// this function is called on local resize only,
 	// remote resize will call this.resize()
 	// need to update the internal geometry
+	this.debug("geometry", "handle_resized(", e, ")");
 	if(e) {
 		this.x = this.x + Math.round(e.position.left - e.originalPosition.left);
 		this.y = this.y + Math.round(e.position.top - e.originalPosition.top);
@@ -677,11 +784,14 @@ XpraWindow.prototype.handle_resized = function(e) {
  * store internal geometry, external is always in CSS left and top
  */
 XpraWindow.prototype.handle_moved = function(e) {
+	var left = Math.round(e.position.left),
+		top = Math.round(e.position.top);
+	this.debug("geometry", "handle_moved(", e, ") left=", left, ", top=", top);
 	// add on padding to the event position so that
 	// it reflects the internal geometry of the canvas
 	//this.log("handle moved: position=", e.position.left, e.position.top);
-	this.x = Math.round(e.position.left) + this.leftoffset;
-	this.y = Math.round(e.position.top) + this.topoffset;
+	this.x = left + this.leftoffset;
+	this.y = top + this.topoffset;
 	// make sure we are visible after move
 	this.ensure_visible();
 	// tell remote we have moved window
@@ -693,10 +803,18 @@ XpraWindow.prototype.handle_moved = function(e) {
  * if it is fullscreen or maximized.
  */
 XpraWindow.prototype.screen_resized = function() {
-	console.log("screen resized");
+	this.debug("geometry", "screen_resized() server_is_desktop=", this.client.server_is_desktop, ", server_is_shadow=", this.client.server_is_shadow);
 	if (this.client.server_is_desktop) {
 		this.match_screen_size();
-		this.handle_resized();
+	}
+	if (this.client.server_is_shadow) {
+		//note: when this window is created,
+		// it may not have been added to the client's list yet
+		var ids = Object.keys(this.client.id_to_window);
+		if (ids.length==0 || ids[0]==this.wid) {
+			//single window, recenter it:
+			this.recenter();
+		}
 	}
 	if (this.fullscreen || this.maximized) {
 		this.fill_screen();
@@ -705,6 +823,28 @@ XpraWindow.prototype.screen_resized = function() {
 	this.ensure_visible();
 };
 
+XpraWindow.prototype.recenter = function(force_update_geometry) {
+	var x = this.x,
+		y = this.y;
+	this.debug("geometry", "recenter() x=", x, ", y=", y, ", desktop size: ", this.client.desktop_width, this.client.desktop_height);
+	x = Math.round((this.client.desktop_width-this.w)/2);
+	y = Math.round((this.client.desktop_height-this.h)/2);
+	if (this.x!=x || this.y!=y || force_update_geometry) {
+		this.debug("geometry", "window re-centered to:", x, y);
+		this.x = x;
+		this.y = y;
+		this.updateCSSGeometry();
+		this.geometry_cb(this);
+	}
+	else {
+		this.debug("geometry", "recenter() unchanged at ", x, y);
+	}
+	if (this.x<0 || this.y<0) {
+		this.warn("window does not fit in canvas, offsets: ", x, y);
+	}
+}
+
+
 XpraWindow.prototype.match_screen_size = function() {
 	var maxw = this.client.desktop_width;
 	var maxh = this.client.desktop_height;
@@ -712,10 +852,11 @@ XpraWindow.prototype.match_screen_size = function() {
 	if (this.client.server_resize_exact) {
 		neww = maxw;
 		newh = maxh;
-		console.log("resizing to exact size:", neww, newh);
+		this.log("resizing to exact size:", neww, newh);
 	}
 	else {
 		if (this.client.server_screen_sizes.length==0) {
+			this.recenter();
 			return;
 		}
 		//try to find the best screen size to use,
@@ -735,7 +876,7 @@ XpraWindow.prototype.match_screen_size = function() {
 			}
 		}
 		if (neww==0 && newh==0) {
-			//not found, try to fine the smallest one:
+			//not found, try to find the smallest one:
 			best = 0;
 			for (var i = 0; i < screen_sizes.length; i++) {
 				screen_size = screen_sizes[i];
@@ -748,9 +889,11 @@ XpraWindow.prototype.match_screen_size = function() {
 				}
 			}
 		}
-		console.log("best screen size:", neww, newh);
+		this.log("best screen size:", neww, newh);
 	}
-	this.resize(neww, newh);
+	this.w = neww;
+	this.h = newh;
+	this.recenter(true);
 };
 
 
@@ -759,6 +902,7 @@ XpraWindow.prototype.match_screen_size = function() {
  */
 
 XpraWindow.prototype.move_resize = function(x, y, w, h) {
+	this.debug("geometry", "move_resize(", x, y, w, h, ")");
 	// only do it if actually changed!
 	if(!(this.w == w) || !(this.h == h) || !(this.x == x) || !(this.y == y)) {
 		this.w = w;
@@ -775,10 +919,12 @@ XpraWindow.prototype.move_resize = function(x, y, w, h) {
 };
 
 XpraWindow.prototype.move = function(x, y) {
+	this.debug("geometry", "move(", x, y, ")");
 	this.move_resize(x, y, this.w, this.h);
 };
 
 XpraWindow.prototype.resize = function(w, h) {
+	this.debug("geometry", "resize(", w, h, ")");
 	this.move_resize(this.x, this.y, w, h);
 };
 
@@ -841,9 +987,17 @@ XpraWindow.prototype.update_icon = function(width, height, encoding, img_data) {
 	if (encoding=="png") {
 		//move title to the right:
 		$("#title"+ String(this.wid)).css('left', 32);
+		if (typeof img_data === 'string') {
+			var uint = new Uint8Array(img_data.length);
+			for(var i=0;i<img_data.length;++i) {
+				uint[i] = img_data.charCodeAt(i);
+			}
+			img_data = uint;
+		}
 		src = "data:image/"+encoding+";base64," + Utilities.ArrayBufferToBase64(img_data);
 	}
 	jQuery('#windowicon' + String(this.wid)).attr('src', src);
+	jQuery('#windowlistitemicon' + String(this.wid)).attr('src', src);
 	return src;
 };
 
@@ -909,10 +1063,10 @@ XpraWindow.prototype._broadway_paint = function(buffer, enc_width, enc_height, w
 	img.data.set(buffer);
 	var x = this.broadway_paint_location[0];
 	var y = this.broadway_paint_location[1];
-	this.offscreen_canvas_ctx.putImageData(img, x, y);
+	this.offscreen_canvas_ctx.putImageData(img, x, y, 0, 0, enc_width, enc_height);
 	if(enc_width!=width || enc_height!=height) {
 		//scale it:
-		this.offscreen_canvas_ctx.drawImage(this.offscreen_canvas, x, y, p_width, p_height, x, y, width, height);
+		this.offscreen_canvas_ctx.drawImage(this.offscreen_canvas, x, y, enc_width, enc_height, x, y, width, height);
 	}
 };
 
@@ -933,7 +1087,7 @@ XpraWindow.prototype._close_video = function() {
 				}
 				this.media_source.endOfStream();
 			} catch(e) {
-				this.warn("video media source EOS error: "+e);
+				this.exc(e, "video media source EOS error");
 			}
 			this.video_source_buffer = null;
 			this.media_source = null;
@@ -1008,7 +1162,7 @@ XpraWindow.prototype._init_video = function(width, height, coding, profile, leve
 		codec_string = 'video/webm;codecs="vp9"';
 	}
 	else {
-		throw Exception("invalid encoding: "+coding);
+		throw new Error("invalid encoding: "+coding);
 	}
 	this.log("video codec string: "+codec_string+" for "+coding+" profile '"+profile+"', level '"+level+"'");
 	this.media_source.addEventListener('sourceopen', function() {
@@ -1122,14 +1276,14 @@ XpraWindow.prototype.do_paint = function paint(x, y, width, height, coding, img_
 	}
 	function paint_box(color, px, py, pw, ph) {
 		me.offscreen_canvas_ctx.strokeStyle = color;
-		me.offscreen_canvas_ctx.lineWidth = "2";
+		me.offscreen_canvas_ctx.lineWidth = 2;
 		me.offscreen_canvas_ctx.strokeRect(px, py, pw, ph);
 	}
 
 	function painted(skip_box) {
 		img_data = null;
 		me.paint_pending = 0;
-		if (me.paint_debug && !skip_box) {
+		if (!skip_box && me.debug_categories.includes("draw")) {
 			var color = DEFAULT_BOX_COLORS[coding] || "white";
 			paint_box(color, x, y, width, height);
 		}
@@ -1306,6 +1460,7 @@ XpraWindow.prototype.do_paint = function paint(x, y, width, height, coding, img_
 		}
 	}
 	catch (e) {
+		this.exc(e, "error painting", coding, "sequence no", packet_sequence);
 		paint_error(e);
 	}
 };
